/*
	Sudoku generator

	Motif interface

	(C) Copyright 2005 by Mogens Kjaer, mk@lemo.dk

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <stdlib.h>
#include <time.h>

#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/PushB.h>
#include <Xm/MainW.h>
#include <Xm/RowColumn.h>
#include <Xm/MessageB.h>
#include <Xm/FileSB.h>
#include <Xm/DialogS.h>
#include <Xm/Separator.h>
#include <Xm/ToggleB.h>
#include <Xm/Protocols.h>
#include <Xm/DrawingA.h>
#include <Xm/ScrolledW.h>
#include <Xm/TextF.h>

static Widget main_wnd=NULL, printer_wnd=NULL, settings_wnd=NULL;
static Widget toplevel;
static Widget main_form, settings_pb, create_pb, solve_pb, print_pb, quit_pb, drawing_wnd;
static Widget main_settingsdialog, settings_form, settings_label, settings_edit, settings_check, settings_symmetrical,
              settings_ok, settings_cancel;
static Widget main_printdialog, printer_fileform, printer_fileselection, papersize_label, papersize,
              papersize_pane, A4_w, Letter_w;

typedef struct
{
  XFontStruct *textfont;
  Pixel redcolor, blackcolor;
} ApplicationData, *ApplicationDataPtr;

static ApplicationData applicationdata;

static XtResource resources[] =
{
  {"textfont", "Textfont", XtRFontStruct, sizeof(XFontStruct *),
   XtOffset(ApplicationDataPtr, textfont),
   XtRString, "-ibm-courier-medium-r-normal--50-*-*-*-m-*-iso8859-1"
  },
  {"redcolor", "Redcolor", XtRPixel, sizeof(int),
   XtOffset(ApplicationDataPtr, redcolor),
   XtRString, "red"
  },
  {"blackcolor", "Blackcolor", XtRPixel, sizeof(int),
   XtOffset(ApplicationDataPtr, blackcolor),
   XtRString, "black"
  },
};

static GC thick_gc, thin_gc, black_gc, red_gc;
static int font_width,font_height;

int nfilled=24;
int newnfilled;
int bestndigits=81;
int symmetrical=1;
static char filled_buffer[4];

#define THINWIDTH 1
#define THICKWIDTH 3
#define CELLSIZE 50


int displayboard[9][9];

#define RNDDIGIT (random()%9+1)
#define NSET 4

int setboard[9][9];
int setboardsave[9][9];
int beforeboard[9][9];
int board[9][9];
int nmove=0;
int nsol=0;
int ntest=0;
int running=0;
int letter=0;
int solved=0;

int n;
Arg wargs[1000];

setcreatemenutext(char *txt)
{
  XmString stmp;

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateSimple(txt))); n++;
  XtSetValues(create_pb, wargs, n);

  XmStringFree(stmp);
}

setsolvemenutext(char *txt)
{
  XmString stmp;

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateSimple(txt))); n++;
  XtSetValues(solve_pb, wargs, n);

  XmStringFree(stmp);
}

settitle(char *txt)
{
  XTextProperty text_prop;

  text_prop.value = (unsigned char *) txt;
  text_prop.encoding = XA_STRING;
  text_prop.format = 8;
  text_prop.nitems = strlen(txt);
  XSetWMName(XtDisplay(toplevel), XtWindow(toplevel), &text_prop);
}

static void generate_sudoku()
{
  bestndigits=81;
  running=1;
  ntest=0;
}

static printer_print()
{
}

static main_clear()
{
  int i,j;
  for(i=0;i<9;i++)
  for(j=0;j<9;j++)
  displayboard[i][j] = 0;
}

main_update()
{
  char txt[16];
  int row,col,size,x,y,width;

  size = 9*CELLSIZE+2*THICKWIDTH+6*THINWIDTH;

  y=0;
  XDrawLine(XtDisplay(drawing_wnd), XtWindow(drawing_wnd), thin_gc,
	    0, 0, size, 0);
  y+=THINWIDTH;
  for(row=0;row<9;row++)
  {
    if(row>0)
    {
      XDrawLine(XtDisplay(drawing_wnd), XtWindow(drawing_wnd), (row%3==0)?thick_gc:thin_gc,
                y, 0, y, size);
      XDrawLine(XtDisplay(drawing_wnd), XtWindow(drawing_wnd), (row%3==0)?thick_gc:thin_gc,
                0, y, size, y);
      y+=(row%3==0)?THICKWIDTH:THINWIDTH;
    }
    x=0;
    for(col=0;col<9;col++)
    {
      if(displayboard[row][col])
      {
        sprintf(txt, "%d", abs(displayboard[row][col]));
      }
      else
      {
        txt[0]=' ';
	txt[1]=0;
      }
      XDrawImageString(XtDisplay(drawing_wnd), XtWindow(drawing_wnd), displayboard[row][col]<0?red_gc:black_gc, x+CELLSIZE/2-font_width/2, y+CELLSIZE-(CELLSIZE-applicationdata.textfont->ascent)/2, txt, strlen(txt));
      if(col>0) x+=(col%3==0)?THICKWIDTH:THINWIDTH;
      x+=CELLSIZE;
    }
    y+=CELLSIZE;
  }
}

static void settings_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  if(!running) XtRealizeWidget(main_settingsdialog);
}

static void create_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  char windowtitle[80];
  if(running)
  {
    setcreatemenutext("Create");
    running=0;
    sprintf(windowtitle, "Sudoku Generator. Stopped at iteration %d, best: %d", ntest, bestndigits);
    settitle(windowtitle);
  }
  else
  {
    setcreatemenutext("Stop!");
    generate_sudoku();
    setsolvemenutext("Solve");
   }
}

static void solve_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  if(!running)
  {
   if(solved)
   {
     unsolve_sudoku();
   }
   else
   {
     solve_sudoku();
   }
  }
}

static void print_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  XtRealizeWidget(main_printdialog);
}

static void quit_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  exit(0);
}

static void drawing_expose_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  main_update();
}

static void symmetrical_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
}

static void settingsok_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  char *s;
  Boolean newsymmetric;

  s = XmTextFieldGetString(settings_edit);
  sscanf(s, "%d", &nfilled);
  XFree(s);

  n=0;
  XtSetArg(wargs[n], XmNset, &newsymmetric); n++;
  XtGetValues(settings_symmetrical, wargs, n);
  symmetrical=newsymmetric;

  main_clear();
  bestndigits=81;
  settitle("Sudoku Generator");
  main_update();

  XtUnrealizeWidget(main_settingsdialog);
}

static void settingscancel_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  XtUnrealizeWidget(main_settingsdialog);
}

static void printer_fileopen_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  XmFileSelectionBoxCallbackStruct *cbs = (XmFileSelectionBoxCallbackStruct *)call_data;
  char *filename;
  FILE *print_fh;
  double dev_xmin, dev_xmax, dev_ymin, dev_ymax, margin;
  char *paper_text;
  double size, printer_cellsize, x, y, fontsize, width, printer_thinwidth, printer_thickwidth;
  int row, col;

  XmStringGetLtoR(cbs->value, XmSTRING_DEFAULT_CHARSET, &filename);

  print_fh = fopen(filename, "w");
  if(print_fh != NULL)
  {
    if(letter)
    {
      dev_xmin = 0;
      dev_xmax = 612;
      dev_ymin = 0;
      dev_ymax = 792;
      margin = 18;
      paper_text = "Letter";
    }
    else
    {
      dev_xmin = 0;
      dev_xmax = 595;
      dev_ymin = 0;
      dev_ymax = 842;
      margin = 18;
      paper_text = "A4";
    }
    
    fprintf(print_fh, "%%!PS-Adobe-3.0\n");
    fprintf(print_fh, "%%%%Creator: Sudoku generator\n");
    fprintf(print_fh, "%%%%Pages: (atend)\n");
    fprintf(print_fh, "%%%%DocumentMedia: %s %.0f %.0f 0 white ()\n", paper_text, dev_xmax, dev_ymax);
    fprintf(print_fh, "%%%%BoundingBox: %.0f %.0f %.0f %.0f\n", dev_xmin, dev_ymin, dev_xmax, dev_ymax);
    fprintf(print_fh, "%%%%Orientation: Portrait\n");
    fprintf(print_fh, "%%%%EndComments\n\n");

    fprintf(print_fh, "[{\n");
    fprintf(print_fh, "%%%%BeginFeature: *PageSize %s\n", paper_text);
    fprintf(print_fh, "<</PageSize[%d %d]/ImagingBBox null>>setpagedevice\n", (int) dev_xmax, (int) dev_ymax);
    fprintf(print_fh, "%%%%EndFeature\n");
    fprintf(print_fh, "} stopped cleartomark\n");

    if((dev_xmax-dev_xmin)>(dev_ymax-dev_ymin))
    {
      size = dev_ymax-dev_ymin;
    }
    else
    {
      size = dev_xmax-dev_xmin;
    }

    size = size-2*margin;
    fontsize = size/10;
    printer_cellsize=size/9;
    printer_thinwidth = size/400;
    printer_thickwidth = printer_thinwidth*3;

    fprintf(print_fh, "%g %g translate\n", (dev_xmax-dev_xmin)/2-size/2, (dev_ymax-dev_ymin)/2-size/2);

    fprintf(print_fh, "/showcenter {dup gsave 0 0 moveto true charpath flattenpath pathbbox grestore /symax exch def /sxmax exch def /symin exch def /sxmin exch def sxmin sxmax sub 2 div symin symax sub 2 div rmoveto show} def\n");
    fprintf(print_fh, "/Courier findfont %g scalefont setfont\n", fontsize);

    y=size;
    for(row=0;row<9;row++)
    {
      if(row>0)
      {
        width = (row%3==0)?printer_thickwidth:printer_thinwidth;
	fprintf(print_fh, "%g setlinewidth\n", width);
	fprintf(print_fh, "0 0 0 setrgbcolor\n");
	fprintf(print_fh, "newpath\n");
	fprintf(print_fh, "%g 0 moveto\n", y);
	fprintf(print_fh, "%g %g lineto\n", y, size);
	fprintf(print_fh, "0 %g moveto\n", y);
	fprintf(print_fh, "%g %g lineto\n", size, y);
	fprintf(print_fh, "stroke\n");
      }
      x=0;
      for(col=0; col<9; col++)
      {
	if(displayboard[row][col])
	{
	  fprintf(print_fh, "%g %g moveto\n", x+printer_cellsize/2,y-printer_cellsize/2);
	  fprintf(print_fh, "%d 0 0 setrgbcolor\n", displayboard[row][col]<0?1:0);
	  fprintf(print_fh, "(%d) showcenter\n", abs(displayboard[row][col]));
	}
	x+=printer_cellsize;
      }
      y-=printer_cellsize;
    }
    fprintf(print_fh, "%g setlinewidth\n", printer_thickwidth);
    fprintf(print_fh, "0 0 0 setrgbcolor\n");
    fprintf(print_fh, "newpath\n");
    fprintf(print_fh, "0 0 moveto\n");
    fprintf(print_fh, "0 %g lineto\n", size);
    fprintf(print_fh, "%g %g lineto\n", size, size);
    fprintf(print_fh, "%g 0 lineto\n", size);
    fprintf(print_fh, "closepath\n");
    fprintf(print_fh, "stroke\n");
    fprintf(print_fh, "showpage\n");
    fprintf(print_fh, "%%%%Pages: 1\n");
    fprintf(print_fh, "%%%%EOF\n%c", (char) 4);

    fclose(print_fh);
  }

  XtFree(filename);
  XtUnrealizeWidget(main_printdialog);
}

static void printer_fileclose_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  XtUnrealizeWidget(main_printdialog);
}

static void printer_papersize_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  letter = (int) client_data;
}

static main_wnd_init()
{
  Atom atom;
  Widget menubar;
  XGCValues values;
  XtGCMask valueMask;

  XmString s[20];
  XmString stmp;

  n=0;
  XtSetArg(wargs[n], XtNtitle, "Sudoku"); n++;
  XtSetArg(wargs[n], XtNgeometry, "+10+10"); n++;
  main_wnd = XmCreateMainWindow(toplevel, "main", wargs, n);
  atom = XmInternAtom(XtDisplay(toplevel),"WM_DELETE_WINDOW",TRUE);
  XmAddWMProtocolCallback(toplevel, atom, quit_cb, False);
  n=0;
  main_form = XtCreateManagedWidget("main_form", xmFormWidgetClass,
                                    main_wnd, wargs, n);

  /* Menu bar */

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  settings_pb = XtCreateManagedWidget("Settings", xmPushButtonWidgetClass,
                                      main_form, wargs, n);
  XtAddCallback(settings_pb, XmNactivateCallback, settings_cb, NULL);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, settings_pb); n++;
  create_pb = XtCreateManagedWidget("Create", xmPushButtonWidgetClass,
                                      main_form, wargs, n);
  XtAddCallback(create_pb, XmNactivateCallback, create_cb, NULL);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, create_pb); n++;
  solve_pb = XtCreateManagedWidget("Solve", xmPushButtonWidgetClass,
                                      main_form, wargs, n);
  XtAddCallback(solve_pb, XmNactivateCallback, solve_cb, NULL);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, solve_pb); n++;
  print_pb = XtCreateManagedWidget("Print", xmPushButtonWidgetClass,
                                      main_form, wargs, n);
  XtAddCallback(print_pb, XmNactivateCallback, print_cb, NULL);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, print_pb); n++;
  quit_pb = XtCreateManagedWidget("Quit", xmPushButtonWidgetClass,
                                      main_form, wargs, n);
  XtAddCallback(quit_pb, XmNactivateCallback, quit_cb, NULL);

  /* Main drawing area */

  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNrightAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNbottomAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNtopWidget, settings_pb); n++;
  XtSetArg(wargs[n], XmNwidth, 9*CELLSIZE+2*THICKWIDTH+6*THINWIDTH); n++;
  XtSetArg(wargs[n], XmNheight, 9*CELLSIZE+2*THICKWIDTH+7*THINWIDTH); n++;
  drawing_wnd = XtCreateManagedWidget("drawing_wnd", xmDrawingAreaWidgetClass,
                                      main_form, wargs, n);
  XtAddCallback(drawing_wnd, XmNexposeCallback, drawing_expose_cb, (XtPointer) NULL);

  
  valueMask = GCForeground | GCBackground | GCLineWidth;
  values.foreground = BlackPixel(XtDisplay(drawing_wnd),XDefaultScreen(XtDisplay(drawing_wnd)));
  values.background = WhitePixel(XtDisplay(drawing_wnd),XDefaultScreen(XtDisplay(drawing_wnd)));
  values.line_width = THINWIDTH;
  thin_gc = XtGetGC(drawing_wnd, valueMask, &values);

  valueMask = GCForeground | GCBackground | GCLineWidth;
  values.foreground = BlackPixel(XtDisplay(drawing_wnd),XDefaultScreen(XtDisplay(drawing_wnd)));
  values.background = WhitePixel(XtDisplay(drawing_wnd),XDefaultScreen(XtDisplay(drawing_wnd)));
  values.line_width = THICKWIDTH;
  thick_gc = XtGetGC(drawing_wnd, valueMask, &values);

  font_width=applicationdata.textfont->max_bounds.width;
  font_height=applicationdata.textfont->ascent+applicationdata.textfont->descent;

  valueMask = GCForeground | GCBackground | GCFont;
  values.foreground = applicationdata.blackcolor;
  values.background = WhitePixel(XtDisplay(drawing_wnd),XDefaultScreen(XtDisplay(drawing_wnd)));
  values.font=applicationdata.textfont->fid;
  black_gc = XtGetGC(drawing_wnd, valueMask, &values);

  valueMask = GCForeground | GCBackground | GCFont;
  values.foreground = applicationdata.redcolor;
  values.background = WhitePixel(XtDisplay(drawing_wnd),XDefaultScreen(XtDisplay(drawing_wnd)));
  values.font=applicationdata.textfont->fid;
  red_gc = XtGetGC(drawing_wnd, valueMask, &values);

  n=0;
  XtSetArg(wargs[n], XtNtitle, "Sudoku settings"); n++;
  main_settingsdialog = XtCreateApplicationShell("sudoku_settings", topLevelShellWidgetClass, wargs, n);

  n=0;
  settings_form = XtCreateManagedWidget("form", xmFormWidgetClass, main_settingsdialog, wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNlabelString, (s[0]=XmStringCreateSimple("No. of filled cells:"))); n++;
  settings_label = XmCreateLabel(settings_form, "FilledTitle", wargs, n);
  XtManageChild(settings_label);
  XmStringFree(s[0]);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftOffset, 10); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, settings_label); n++;
  XtSetArg(wargs[n], XmNvalue, filled_buffer); n++;
  sprintf(filled_buffer, "%d", nfilled);
  XtSetArg(wargs[n], XmNmaxLength, sizeof(filled_buffer)-1); n++;
  XtSetArg(wargs[n], XmNcolumns, sizeof(filled_buffer)-1); n++;
  XtSetArg(wargs[n], XmNverifyBell, False); n++;

  settings_edit = XtCreateManagedWidget("settings_edit", xmTextFieldWidgetClass, settings_form, wargs, n);
  XtManageChild(settings_edit);

  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, settings_label); n++;
  settings_check = XmCreateSimpleCheckBox(settings_form, "CheckBox1", wargs, n);
  XtManageChild(settings_check);

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (s[0]=XmStringCreateSimple("Symmetrical"))); n++;
  XtSetArg(wargs[n], XmNset, symmetrical?XmSET:XmUNSET); n++;
  settings_symmetrical = XmCreateToggleButton(settings_check, "ToggleBuffer", wargs, n);
  XtAddCallback(settings_symmetrical, XmNvalueChangedCallback, symmetrical_cb, NULL);
  XtManageChild(settings_symmetrical);
  XmStringFree(s[0]);
  
  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftOffset, 10); n++;
  XtSetArg(wargs[n], XmNbottomAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, settings_check); n++;
  XtSetArg(wargs[n], XmNtopOffset, 10); n++;
  XtSetArg(wargs[n], XmNbottomOffset, 10); n++;
  settings_ok = XtCreateManagedWidget("OK", xmPushButtonWidgetClass, settings_form, wargs, n);
  XtAddCallback(settings_ok, XmNactivateCallback, settingsok_cb, NULL);

  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, settings_ok); n++;
  XtSetArg(wargs[n], XmNleftOffset, 10); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, settings_ok); n++;
  settings_cancel = XtCreateManagedWidget("Cancel", xmPushButtonWidgetClass, settings_form, wargs, n);
  XtAddCallback(settings_cancel, XmNactivateCallback, settingscancel_cb, NULL);

  n=0;
  XtSetArg(wargs[n], XtNtitle, "PostScript File Selection"); n++;
  main_printdialog = XtCreateApplicationShell("main_printdialog", topLevelShellWidgetClass, wargs, n);

  n=0;
  printer_fileform = XtCreateManagedWidget("form", xmFormWidgetClass, main_printdialog, wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNpattern, (stmp=XmStringCreateSimple("*.ps"))); n++;
  printer_fileselection = XtCreateManagedWidget("printer_fileselection", xmFileSelectionBoxWidgetClass, printer_fileform, wargs, n);
  XmStringFree(stmp);

  XtAddCallback(printer_fileselection, XmNokCallback, printer_fileopen_cb, (XtPointer)NULL);
  XtAddCallback(printer_fileselection, XmNcancelCallback, printer_fileclose_cb, (XtPointer)NULL);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNtopOffset, 5); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, printer_fileselection); n++;
  XtSetArg(wargs[n], XmNleftOffset, 10); n++;
  papersize_label = XtCreateManagedWidget("Select paper size", xmLabelWidgetClass, printer_fileform, wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, papersize_label); n++;
  XtSetArg(wargs[n], XmNleftOffset, 0); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, printer_fileselection); n++;
  papersize = XmCreateOptionMenu(printer_fileform, "papersize", wargs, n);

  n=0;
  papersize_pane = XmCreatePulldownMenu(papersize, "pulldown", wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateLocalized("A4"))); n++;
  A4_w = XtCreateManagedWidget("A4_select", xmPushButtonWidgetClass, papersize_pane, wargs, n);
  XtAddCallback(A4_w, XmNactivateCallback, printer_papersize_cb, (XtPointer) 0);

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateLocalized("Letter"))); n++;
  Letter_w = XtCreateManagedWidget("letter_select", xmPushButtonWidgetClass, papersize_pane, wargs, n);
  XtAddCallback(Letter_w, XmNactivateCallback, printer_papersize_cb, (XtPointer) 1);

  n=0;
  XtSetArg(wargs[n], XmNsubMenuId, papersize_pane); n++;
  XtSetValues(papersize, wargs, n);

  XtManageChild(papersize);

  n=0;
  XtSetArg(wargs[n], XmNmenuHistory, letter?Letter_w:A4_w); n++;
  XtSetValues(papersize, wargs, n);


  XmFileSelectionDoSearch(printer_fileselection, NULL);
  XtManageChild(main_wnd);
}

static void sudokumain()
{
  XEvent event;
  int status;

AGAIN:
  status = XtPending();
  if(status)
  {
AGAIN2:
    XtNextEvent(&event);
    XtDispatchEvent(&event);
    goto AGAIN;
  }
  if(running)
  {
    sudoku_run();
    goto AGAIN;
  }
  goto AGAIN2;
}  

main(int argc, char **argv)
{

  srandom(time(NULL));

  toplevel=XtInitialize("Sudoku", "Sudoku", NULL, 0, &argc, argv);

  XtGetApplicationResources(toplevel, &applicationdata,
                            resources, XtNumber(resources), NULL, 0);

  main_wnd_init();

  running = 0;

  XtRealizeWidget(toplevel);

  settitle("Sudoku generator");

  sudokumain();
}
